@@ -1,11 +1,21 @@
 import pytest
 
 from hatchling.metadata.core import BuildMetadata, CoreMetadata, HatchMetadata, ProjectMetadata
+from hatchling.metadata.spec import (
+    LATEST_METADATA_VERSION,
+    get_core_metadata_constructors,
+    project_metadata_from_core_metadata,
+)
 from hatchling.plugin.manager import PluginManager
 from hatchling.utils.constants import DEFAULT_BUILD_SCRIPT
 from hatchling.version.source.regex import RegexSource
 
 
+@pytest.fixture(scope='module')
+def latest_spec():
+    return get_core_metadata_constructors()[LATEST_METADATA_VERSION]
+
+
 class TestConfig:
     def test_default(self, isolation):
         metadata = ProjectMetadata(str(isolation), None)
@@ -32,13 +42,8 @@ class TestInterface:
     def test_types(self, isolation):
         metadata = ProjectMetadata(str(isolation), None, {'project': {}})
 
-        assert metadata.core is metadata._core
         assert isinstance(metadata.core, CoreMetadata)
-
-        assert metadata.hatch is metadata._hatch
         assert isinstance(metadata.hatch, HatchMetadata)
-
-        assert metadata.build is metadata._build
         assert isinstance(metadata.build, BuildMetadata)
 
     def test_missing_core_metadata(self, isolation):
@@ -89,9 +94,51 @@ class TestDynamic:
         dynamic = ['version']
         metadata = ProjectMetadata(str(isolation), None, {'project': {'dynamic': dynamic}})
 
-        assert metadata.core.dynamic is dynamic
         assert metadata.core.dynamic == ['version']
 
+    def test_cache_not_array(self, isolation):
+        metadata = ProjectMetadata(str(isolation), None, {'project': {'dynamic': 10}})
+
+        with pytest.raises(TypeError, match='Field `project.dynamic` must be an array'):
+            _ = metadata.dynamic
+
+    def test_cache_entry_not_string(self, isolation):
+        metadata = ProjectMetadata(str(isolation), None, {'project': {'dynamic': [10]}})
+
+        with pytest.raises(TypeError, match='Field #1 of field `project.dynamic` must be a string'):
+            _ = metadata.dynamic
+
+    def test_cache_correct(self, temp_dir, helpers):
+        metadata = ProjectMetadata(
+            str(temp_dir),
+            PluginManager(),
+            {
+                'project': {'name': 'foo', 'dynamic': ['version', 'description']},
+                'tool': {'hatch': {'version': {'path': 'a/b'}, 'metadata': {'hooks': {'custom': {}}}}},
+            },
+        )
+
+        file_path = temp_dir / 'a' / 'b'
+        file_path.ensure_parent_dir_exists()
+        file_path.write_text('__version__ = "0.0.1"')
+
+        file_path = temp_dir / DEFAULT_BUILD_SCRIPT
+        file_path.write_text(
+            helpers.dedent(
+                """
+                from hatchling.metadata.plugin.interface import MetadataHookInterface
+
+                class CustomHook(MetadataHookInterface):
+                    def update(self, metadata):
+                        metadata['description'] = metadata['name'] + 'bar'
+                """
+            )
+        )
+
+        # Trigger hooks with `metadata.core` first
+        assert metadata.core.dynamic == []
+        assert metadata.dynamic == ['version', 'description']
+
 
 class TestRawName:
     def test_dynamic(self, isolation):
@@ -295,6 +342,11 @@ class TestDescription:
 
         assert metadata.core.description == metadata.core.description == 'foo'
 
+    def test_normaliza(self, isolation):
+        metadata = ProjectMetadata(str(isolation), None, {'project': {'description': '\nfirst line.\r\nsecond line'}})
+
+        assert metadata.core.description == metadata.core.description == ' first line. second line'
+
 
 class TestReadme:
     def test_dynamic(self, isolation):
@@ -335,7 +387,7 @@ class TestReadme:
             _ = metadata.core.readme
 
     @pytest.mark.parametrize(
-        'extension, content_type', [('.md', 'text/markdown'), ('.rst', 'text/x-rst'), ('.txt', 'text/plain')]
+        ('extension', 'content_type'), [('.md', 'text/markdown'), ('.rst', 'text/x-rst'), ('.txt', 'text/plain')]
     )
     def test_string_correct(self, extension, content_type, temp_dir):
         metadata = ProjectMetadata(str(temp_dir), None, {'project': {'readme': f'foo/bar{extension}'}})
@@ -503,6 +555,12 @@ class TestLicense:
 
         assert metadata.core.license_expression == 'MIT OR Apache-2.0'
 
+    def test_invalid_expression(self, isolation):
+        metadata = ProjectMetadata(str(isolation), None, {'project': {'license': 'mit or foo'}})
+
+        with pytest.raises(ValueError, match="Error parsing field `project.license` - Unknown license: 'foo'"):
+            _ = metadata.core.license_expression
+
     def test_multiple_options(self, isolation):
         metadata = ProjectMetadata(str(isolation), None, {'project': {'license': {'file': '', 'text': ''}}})
 
@@ -563,54 +621,16 @@ class TestLicenseFiles:
         ):
             _ = metadata.core.license_files
 
-    def test_not_table(self, isolation):
+    def test_not_array(self, isolation):
         metadata = ProjectMetadata(str(isolation), None, {'project': {'license-files': 9000}})
 
-        with pytest.raises(TypeError, match='Field `project.license-files` must be a table'):
+        with pytest.raises(TypeError, match='Field `project.license-files` must be an array'):
             _ = metadata.core.license_files
 
-    def test_multiple_options(self, isolation):
-        metadata = ProjectMetadata(str(isolation), None, {'project': {'license-files': {'paths': [], 'globs': []}}})
+    def test_entry_not_string(self, isolation):
+        metadata = ProjectMetadata(str(isolation), None, {'project': {'license-files': [9000]}})
 
-        with pytest.raises(
-            ValueError, match='Cannot specify both `paths` and `globs` in the `project.license-files` table'
-        ):
-            _ = metadata.core.license_files
-
-    def test_no_option(self, isolation):
-        metadata = ProjectMetadata(str(isolation), None, {'project': {'license-files': {}}})
-
-        with pytest.raises(
-            ValueError, match='Must specify either `paths` or `globs` in the `project.license-files` table if defined'
-        ):
-            _ = metadata.core.license_files
-
-    def test_paths_not_array(self, isolation):
-        metadata = ProjectMetadata(str(isolation), None, {'project': {'license-files': {'paths': 9000}}})
-
-        with pytest.raises(TypeError, match='Field `paths` in the `project.license-files` table must be an array'):
-            _ = metadata.core.license_files
-
-    def test_paths_entry_not_string(self, isolation):
-        metadata = ProjectMetadata(str(isolation), None, {'project': {'license-files': {'paths': [9000]}}})
-
-        with pytest.raises(
-            TypeError, match='Entry #1 in field `paths` in the `project.license-files` table must be a string'
-        ):
-            _ = metadata.core.license_files
-
-    def test_globs_not_array(self, isolation):
-        metadata = ProjectMetadata(str(isolation), None, {'project': {'license-files': {'globs': 9000}}})
-
-        with pytest.raises(TypeError, match='Field `globs` in the `project.license-files` table must be an array'):
-            _ = metadata.core.license_files
-
-    def test_globs_entry_not_string(self, isolation):
-        metadata = ProjectMetadata(str(isolation), None, {'project': {'license-files': {'globs': [9000]}}})
-
-        with pytest.raises(
-            TypeError, match='Entry #1 in field `globs` in the `project.license-files` table must be a string'
-        ):
+        with pytest.raises(TypeError, match='Entry #1 of field `project.license-files` must be a string'):
             _ = metadata.core.license_files
 
     def test_default_globs_no_licenses(self, isolation):
@@ -635,7 +655,7 @@ class TestLicenseFiles:
         assert metadata.core.license_files == sorted(expected)
 
     def test_globs_with_licenses(self, temp_dir):
-        metadata = ProjectMetadata(str(temp_dir), None, {'project': {'license-files': {'globs': ['LICENSES/*']}}})
+        metadata = ProjectMetadata(str(temp_dir), None, {'project': {'license-files': ['LICENSES/*']}})
 
         licenses_dir = temp_dir / 'LICENSES'
         licenses_dir.mkdir()
@@ -651,7 +671,7 @@ class TestLicenseFiles:
         metadata = ProjectMetadata(
             str(temp_dir),
             None,
-            {'project': {'license-files': {'paths': ['LICENSES/Apache-2.0.txt', 'LICENSES/MIT.txt', 'COPYING']}}},
+            {'project': {'license-files': ['LICENSES/Apache-2.0.txt', 'LICENSES/MIT.txt', 'COPYING']}},
         )
 
         licenses_dir = temp_dir / 'LICENSES'
@@ -664,20 +684,6 @@ class TestLicenseFiles:
 
         assert metadata.core.license_files == ['COPYING', 'LICENSES/Apache-2.0.txt', 'LICENSES/MIT.txt']
 
-    def test_paths_missing_license(self, temp_dir):
-        metadata = ProjectMetadata(
-            str(temp_dir),
-            None,
-            {'project': {'license-files': {'paths': ['LICENSES/MIT.txt']}}},
-        )
-
-        licenses_dir = temp_dir / 'LICENSES'
-        licenses_dir.mkdir()
-        (licenses_dir / 'Apache-2.0.txt').touch()
-
-        with pytest.raises(OSError, match='License file does not exist: LICENSES/MIT.txt'):
-            _ = metadata.core.license_files
-
 
 class TestAuthors:
     def test_dynamic(self, isolation):
@@ -892,10 +898,57 @@ class TestClassifiers:
         with pytest.raises(TypeError, match='Classifier #1 of field `project.classifiers` must be a string'):
             _ = metadata.core.classifiers
 
-    def test_correct(self, isolation):
-        metadata = ProjectMetadata(str(isolation), None, {'project': {'classifiers': ['foo', 'foo', 'bar']}})
+    def test_entry_unknown(self, isolation, monkeypatch):
+        monkeypatch.delenv('HATCH_METADATA_CLASSIFIERS_NO_VERIFY', False)
+        metadata = ProjectMetadata(str(isolation), None, {'project': {'classifiers': ['foo']}})
 
-        assert metadata.core.classifiers == metadata.core.classifiers == ['bar', 'foo']
+        with pytest.raises(ValueError, match='Unknown classifier in field `project.classifiers`: foo'):
+            _ = metadata.core.classifiers
+
+    def test_entry_unknown_no_verify(self, isolation, monkeypatch):
+        monkeypatch.setenv('HATCH_METADATA_CLASSIFIERS_NO_VERIFY', '1')
+        classifiers = [
+            'Programming Language :: Python :: 3.11',
+            'Programming Language :: Python :: 3.11',
+            'Programming Language :: Python :: 3.9',
+            'Development Status :: 4 - Beta',
+            'Private :: Do Not Upload',
+            'Foo',
+        ]
+        metadata = ProjectMetadata(str(isolation), None, {'project': {'classifiers': classifiers}})
+
+        assert (
+            metadata.core.classifiers
+            == metadata.core.classifiers
+            == [
+                'Private :: Do Not Upload',
+                'Development Status :: 4 - Beta',
+                'Foo',
+                'Programming Language :: Python :: 3.9',
+                'Programming Language :: Python :: 3.11',
+            ]
+        )
+
+    def test_correct(self, isolation):
+        classifiers = [
+            'Programming Language :: Python :: 3.11',
+            'Programming Language :: Python :: 3.11',
+            'Programming Language :: Python :: 3.9',
+            'Development Status :: 4 - Beta',
+            'Private :: Do Not Upload',
+        ]
+        metadata = ProjectMetadata(str(isolation), None, {'project': {'classifiers': classifiers}})
+
+        assert (
+            metadata.core.classifiers
+            == metadata.core.classifiers
+            == [
+                'Private :: Do Not Upload',
+                'Development Status :: 4 - Beta',
+                'Programming Language :: Python :: 3.9',
+                'Programming Language :: Python :: 3.11',
+            ]
+        )
 
 
 class TestURLs:
@@ -1004,14 +1057,15 @@ class TestEntryPoints:
         with pytest.raises(TypeError, match='Field `project.entry-points` must be a table'):
             _ = metadata.core.entry_points
 
-    @pytest.mark.parametrize('field', ['scripts', 'gui-scripts'])
-    def test_forbidden_fields(self, isolation, field):
+    @pytest.mark.parametrize(('field', 'expected'), [('console_scripts', 'scripts'), ('gui-scripts', 'gui-scripts')])
+    def test_forbidden_fields(self, isolation, field, expected):
         metadata = ProjectMetadata(str(isolation), None, {'project': {'entry-points': {field: 'foo'}}})
 
         with pytest.raises(
             ValueError,
             match=(
-                f'Field `{field}` must be defined as `project.{field}` instead of in the `project.entry-points` table'
+                f'Field `{field}` must be defined as `project.{expected}` instead of '
+                f'in the `project.entry-points` table'
             ),
         ):
             _ = metadata.core.entry_points
@@ -1092,7 +1146,11 @@ class TestDependencies:
         )
 
         with pytest.raises(
-            ValueError, match='Dependency #1 of field `project.dependencies` cannot be a direct reference'
+            ValueError,
+            match=(
+                'Dependency #1 of field `project.dependencies` cannot be a direct reference unless '
+                'field `tool.hatch.metadata.allow-direct-references` is set to `true`'
+            ),
         ):
             _ = metadata.core.dependencies
 
@@ -1221,6 +1279,44 @@ class TestOptionalDependencies:
         ):
             _ = metadata.core.optional_dependencies
 
+    def test_recursive_circular(self, isolation):
+        metadata = ProjectMetadata(
+            str(isolation),
+            None,
+            {'project': {'name': 'my-app', 'optional-dependencies': {'foo': ['my-app[bar]'], 'bar': ['my-app[foo]']}}},
+        )
+
+        with pytest.raises(
+            ValueError,
+            match='Field `project.optional-dependencies` defines a circular dependency group: foo',
+        ):
+            _ = metadata.core.optional_dependencies
+
+    def test_recursive_unknown(self, isolation):
+        metadata = ProjectMetadata(
+            str(isolation),
+            None,
+            {'project': {'name': 'my-app', 'optional-dependencies': {'foo': ['my-app[bar]']}}},
+        )
+
+        with pytest.raises(
+            ValueError,
+            match='Unknown recursive dependency group in field `project.optional-dependencies`: bar',
+        ):
+            _ = metadata.core.optional_dependencies
+
+    def test_allow_ambiguity(self, isolation):
+        metadata = ProjectMetadata(
+            str(isolation),
+            None,
+            {
+                'project': {'optional-dependencies': {'foo_bar': [], 'foo.bar': []}},
+                'tool': {'hatch': {'metadata': {'allow-ambiguous-features': True}}},
+            },
+        )
+
+        assert metadata.core.optional_dependencies == {'foo_bar': [], 'foo.bar': []}
+
     def test_direct_reference(self, isolation):
         metadata = ProjectMetadata(
             str(isolation),
@@ -1230,7 +1326,10 @@ class TestOptionalDependencies:
 
         with pytest.raises(
             ValueError,
-            match='Dependency #1 of option `foo` of field `project.optional-dependencies` cannot be a direct reference',
+            match=(
+                'Dependency #1 of option `foo` of field `project.optional-dependencies` cannot be a direct reference '
+                'unless field `tool.hatch.metadata.allow-direct-references` is set to `true`'
+            ),
         ):
             _ = metadata.core.optional_dependencies
 
@@ -1239,7 +1338,7 @@ class TestOptionalDependencies:
             str(isolation),
             None,
             {
-                'project': {'optional-dependencies': {'foo': ['proj @ {root:uri}']}},
+                'project': {'name': 'my-app', 'optional-dependencies': {'foo': ['proj @ {root:uri}']}},
                 'tool': {'hatch': {'metadata': {'allow-direct-references': True}}},
             },
         )
@@ -1252,7 +1351,10 @@ class TestOptionalDependencies:
             str(isolation),
             None,
             {
-                'project': {'optional-dependencies': {'foo': ['proj @ git+https://github.com/org/proj.git@v1']}},
+                'project': {
+                    'name': 'my-app',
+                    'optional-dependencies': {'foo': ['proj @ git+https://github.com/org/proj.git@v1']},
+                },
                 'tool': {'hatch': {'metadata': {'allow-direct-references': True}}},
             },
         )
@@ -1265,14 +1367,19 @@ class TestOptionalDependencies:
             None,
             {
                 'project': {
+                    'name': 'my-app',
                     'optional-dependencies': {
                         'foo': [
                             'python___dateutil;platform_python_implementation=="CPython"',
                             'bAr.Baz[TLS, Zu.Bat, EdDSA, Zu_Bat]   >=1.2RC5 , <9000B1',
                             'Foo;python_version<"3.8"',
                             'fOO;     python_version<    "3.8"',
+                            'MY-APP[zZz]',
                         ],
                         'bar': ['foo', 'bar', 'Baz'],
+                        'baz': ['my___app[XYZ]'],
+                        'xyz': ['my...app[Bar]'],
+                        'zzz': ['aaa'],
                     },
                 },
             },
@@ -1283,11 +1390,15 @@ class TestOptionalDependencies:
             == metadata.core.optional_dependencies
             == {
                 'bar': ['bar', 'baz', 'foo'],
+                'baz': ['bar', 'baz', 'foo'],
                 'foo': [
+                    'aaa',
                     'bar-baz[eddsa,tls,zu-bat]<9000b1,>=1.2rc5',
                     "foo; python_version < '3.8'",
                     "python-dateutil; platform_python_implementation == 'CPython'",
                 ],
+                'xyz': ['bar', 'baz', 'foo'],
+                'zzz': ['aaa'],
             }
         )
 
@@ -1304,11 +1415,19 @@ class TestHook:
             _ = metadata.core
 
     def test_custom(self, temp_dir, helpers):
+        classifiers = [
+            'Programming Language :: Python :: 3.11',
+            'Programming Language :: Python :: 3.11',
+            'Programming Language :: Python :: 3.9',
+            'Framework :: Foo',
+            'Development Status :: 4 - Beta',
+            'Private :: Do Not Upload',
+        ]
         metadata = ProjectMetadata(
             str(temp_dir),
             PluginManager(),
             {
-                'project': {'name': 'foo', 'dynamic': ['version', 'description']},
+                'project': {'name': 'foo', 'classifiers': classifiers, 'dynamic': ['version', 'description']},
                 'tool': {'hatch': {'version': {'path': 'a/b'}, 'metadata': {'hooks': {'custom': {}}}}},
             },
         )
@@ -1327,6 +1446,9 @@ class TestHook:
                     def update(self, metadata):
                         metadata['description'] = metadata['name'] + 'bar'
                         metadata['version'] = metadata['version'] + 'rc0'
+
+                    def get_known_classifiers(self):
+                        return ['Framework :: Foo']
                 """
             )
         )
@@ -1335,6 +1457,13 @@ class TestHook:
         assert metadata.core.name == 'foo'
         assert metadata.core.description == 'foobar'
         assert metadata.core.version == '0.0.1rc0'
+        assert metadata.core.classifiers == [
+            'Private :: Do Not Upload',
+            'Development Status :: 4 - Beta',
+            'Framework :: Foo',
+            'Programming Language :: Python :: 3.9',
+            'Programming Language :: Python :: 3.11',
+        ]
 
     def test_custom_missing_dynamic(self, temp_dir, helpers):
         metadata = ProjectMetadata(
@@ -1371,7 +1500,7 @@ class TestHook:
 
 
 class TestHatchPersonalProjectConfigFile:
-    def test_correct(self, isolation, temp_dir, helpers):
+    def test_correct(self, temp_dir, helpers):
         metadata = ProjectMetadata(
             str(temp_dir),
             PluginManager(),
@@ -1399,7 +1528,7 @@ class TestHatchPersonalProjectConfigFile:
         assert metadata.version == '0.0.1'
         assert metadata.hatch.build_config['reproducible'] is False
 
-    def test_precedence(self, isolation, temp_dir, helpers):
+    def test_precedence(self, temp_dir, helpers):
         metadata = ProjectMetadata(
             str(temp_dir),
             PluginManager(),
@@ -1430,3 +1559,226 @@ class TestHatchPersonalProjectConfigFile:
 
         assert metadata.version == '0.0.2'
         assert metadata.hatch.build_config['reproducible'] is False
+
+
+class TestMetadataConversion:
+    def test_required_only(self, isolation, latest_spec):
+        raw_metadata = {'name': 'My.App', 'version': '0.0.1'}
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_dynamic(self, isolation, latest_spec):
+        raw_metadata = {'name': 'My.App', 'version': '0.0.1', 'dynamic': ['authors', 'classifiers']}
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_description(self, isolation, latest_spec):
+        raw_metadata = {'name': 'My.App', 'version': '0.0.1', 'description': 'foo bar'}
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_urls(self, isolation, latest_spec):
+        raw_metadata = {'name': 'My.App', 'version': '0.0.1', 'urls': {'foo': 'bar', 'bar': 'baz'}}
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_authors(self, isolation, latest_spec):
+        raw_metadata = {
+            'name': 'My.App',
+            'version': '0.0.1',
+            'authors': [{'name': 'foobar'}, {'email': 'bar@domain', 'name': 'foo'}, {'email': 'baz@domain'}],
+        }
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_maintainers(self, isolation, latest_spec):
+        raw_metadata = {
+            'name': 'My.App',
+            'version': '0.0.1',
+            'maintainers': [{'name': 'foobar'}, {'email': 'bar@domain', 'name': 'foo'}, {'email': 'baz@domain'}],
+        }
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_keywords(self, isolation, latest_spec):
+        raw_metadata = {'name': 'My.App', 'version': '0.0.1', 'keywords': ['bar', 'foo']}
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_classifiers(self, isolation, latest_spec):
+        raw_metadata = {
+            'name': 'My.App',
+            'version': '0.0.1',
+            'classifiers': ['Programming Language :: Python :: 3.9', 'Programming Language :: Python :: 3.11'],
+        }
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_license_files(self, temp_dir, latest_spec):
+        raw_metadata = {
+            'name': 'My.App',
+            'version': '0.0.1',
+            'license-files': ['LICENSES/Apache-2.0.txt', 'LICENSES/MIT.txt'],
+        }
+        metadata = ProjectMetadata(str(temp_dir), None, {'project': raw_metadata})
+
+        licenses_path = temp_dir / 'LICENSES'
+        licenses_path.mkdir()
+        licenses_path.joinpath('Apache-2.0.txt').touch()
+        licenses_path.joinpath('MIT.txt').touch()
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_license_expression(self, isolation, latest_spec):
+        raw_metadata = {'name': 'My.App', 'version': '0.0.1', 'license': 'MIT'}
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_license_legacy(self, isolation, latest_spec):
+        raw_metadata = {'name': 'My.App', 'version': '0.0.1', 'license': {'text': 'foo'}}
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_readme(self, isolation, latest_spec):
+        raw_metadata = {
+            'name': 'My.App',
+            'version': '0.0.1',
+            'readme': {'content-type': 'text/markdown', 'text': 'test content\n'},
+        }
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_requires_python(self, isolation, latest_spec):
+        raw_metadata = {'name': 'My.App', 'version': '0.0.1', 'requires-python': '<2,>=1'}
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+    def test_dependencies(self, isolation, latest_spec):
+        raw_metadata = {
+            'name': 'My.App',
+            'version': '0.0.1',
+            'dependencies': ['bar==5', 'foo==1'],
+            'optional-dependencies': {
+                'feature1': ['bar==5; python_version < "3"', 'foo==1'],
+                'feature2': ['bar==5', 'foo==1; python_version < "3"'],
+            },
+        }
+        metadata = ProjectMetadata(str(isolation), None, {'project': raw_metadata})
+
+        core_metadata = latest_spec(metadata)
+        assert project_metadata_from_core_metadata(core_metadata) == raw_metadata
+
+
+class TestSourceDistributionMetadata:
+    def test_basic_persistence(self, temp_dir, helpers):
+        metadata = ProjectMetadata(
+            str(temp_dir),
+            None,
+            {
+                'project': {
+                    'name': 'My.App',
+                    'dynamic': ['version', 'keywords'],
+                    'dependencies': ['foo==1'],
+                    'scripts': {'foo': 'bar'},
+                },
+            },
+        )
+
+        pkg_info = temp_dir / 'PKG-INFO'
+        pkg_info.write_text(
+            helpers.dedent(
+                f"""
+                Metadata-Version: {LATEST_METADATA_VERSION}
+                Name: My.App
+                Version: 0.0.1
+                Keywords: foo,bar
+                Requires-Dist: bar==5
+                """
+            )
+        )
+        with temp_dir.as_cwd():
+            assert metadata.core.config == {
+                'name': 'My.App',
+                'version': '0.0.1',
+                'dependencies': ['foo==1'],
+                'keywords': ['foo', 'bar'],
+                'scripts': {'foo': 'bar'},
+                'dynamic': [],
+            }
+
+    def test_metadata_hooks(self, temp_dir, helpers):
+        metadata = ProjectMetadata(
+            str(temp_dir),
+            PluginManager(),
+            {
+                'project': {
+                    'name': 'My.App',
+                    'dynamic': ['version', 'keywords', 'description', 'scripts'],
+                    'dependencies': ['foo==1'],
+                },
+                'tool': {'hatch': {'metadata': {'hooks': {'custom': {}}}}},
+            },
+        )
+
+        build_script = temp_dir / DEFAULT_BUILD_SCRIPT
+        build_script.write_text(
+            helpers.dedent(
+                """
+                from hatchling.metadata.plugin.interface import MetadataHookInterface
+
+                class CustomHook(MetadataHookInterface):
+                    def update(self, metadata):
+                        metadata['description'] = metadata['name'] + ' bar'
+                        metadata['scripts'] = {'foo': 'bar'}
+                """
+            )
+        )
+
+        pkg_info = temp_dir / 'PKG-INFO'
+        pkg_info.write_text(
+            helpers.dedent(
+                f"""
+                Metadata-Version: {LATEST_METADATA_VERSION}
+                Name: My.App
+                Version: 0.0.1
+                Summary: My.App bar
+                Keywords: foo,bar
+                Requires-Dist: bar==5
+                """
+            )
+        )
+        with temp_dir.as_cwd():
+            assert metadata.core.config == {
+                'name': 'My.App',
+                'version': '0.0.1',
+                'description': 'My.App bar',
+                'dependencies': ['foo==1'],
+                'keywords': ['foo', 'bar'],
+                'scripts': {'foo': 'bar'},
+                'dynamic': ['scripts'],
+            }
