@@ -1,40 +1,33 @@
 import os
 import secrets
 import tarfile
-import time
 import zipfile
 from collections import defaultdict
 
-import httpx
 import pytest
 
 from hatch.config.constants import PublishEnvVars
-from hatch.utils.ci import running_in_ci
-
-PUBLISHER_TOKEN = os.environ.get('HATCH_CI_PUBLISHER_TOKEN')
 
 pytestmark = [
-    pytest.mark.skipif(not PUBLISHER_TOKEN, reason='Publishing tests are only executed within CI environments'),
-    pytest.mark.xfail(reason='Test PyPI can be flaky'),
+    pytest.mark.requires_docker,
+    pytest.mark.requires_internet,
+    pytest.mark.usefixtures('devpi'),
+    pytest.mark.usefixtures('mock_backend_process'),
 ]
 
 
-@pytest.fixture(autouse=True)
-def local_builder(mock_backend_process, mocker):
-    if mock_backend_process:
-        mocker.patch('hatch.env.virtual.VirtualEnvironment.build_environment')
-
-    yield
-
-
 @pytest.fixture(autouse=True)
 def keyring_store(mocker):
     mock_store = defaultdict(dict)
-    mocker.patch('keyring.get_password', side_effect=lambda system, user: mock_store[system].get(user))
     mocker.patch(
-        'keyring.set_password', side_effect=lambda system, user, auth: mock_store[system].__setitem__(user, auth)
+        'keyring.get_password',
+        side_effect=lambda system, user: mock_store[system].get(user),
     )
-    yield mock_store
+    mocker.patch(
+        'keyring.set_password',
+        side_effect=lambda system, user, auth: mock_store[system].__setitem__(user, auth),
+    )
+    return mock_store
 
 
 @pytest.fixture
@@ -42,28 +35,6 @@ def published_project_name():
     return f'c4880cdbe05de9a28415fbad{secrets.choice(range(100))}'
 
 
-def wait_for_artifacts(project_name, *artifacts):
-    index_url = f'https://test.pypi.org/simple/{project_name}/'
-    artifact_names = [artifact.name for artifact in artifacts]
-
-    for _ in range(120):
-        try:
-            response = httpx.get(index_url)
-            response.raise_for_status()
-        except Exception:  # no cov
-            pass
-        else:
-            for artifact_name in artifact_names:
-                if artifact_name not in response.text:
-                    break
-            else:
-                break
-
-        time.sleep(1)
-    else:  # no cov
-        raise Exception(f'Could not find artifacts at {index_url}: {", ".join(artifact_names)}')
-
-
 def remove_metadata_field(field: str, metadata_file_contents: str):
     lines = metadata_file_contents.splitlines(True)
 
@@ -86,8 +57,8 @@ def timestamp_to_version(timestamp):
         normalized_minor = str(int(minor))
         padding = '.'.join('0' for _ in range(len(minor) - len(normalized_minor)))
         return f'{major}.{padding}.{normalized_minor}'
-    else:
-        return f'{major}.{minor}'
+
+    return f'{major}.{minor}'
 
 
 def test_timestamp_to_version():
@@ -149,6 +120,44 @@ def test_disabled(hatch, temp_dir, config_file):
     assert result.output == 'Publisher is disabled: index\n'
 
 
+def test_repo_invalid_type(hatch, temp_dir, config_file):
+    config_file.model.publish['index']['repos'] = {'dev': 9000}
+    config_file.save()
+
+    project_name = 'My.App'
+
+    with temp_dir.as_cwd():
+        result = hatch('new', project_name)
+        assert result.exit_code == 0, result.output
+
+    path = temp_dir / 'my-app'
+
+    with path.as_cwd():
+        result = hatch('publish', '--user', 'foo', '--auth', 'bar')
+
+    assert result.exit_code == 1, result.output
+    assert result.output == 'Hatch config field `publish.index.repos.dev` must be a string or a mapping\n'
+
+
+def test_repo_missing_url(hatch, temp_dir, config_file):
+    config_file.model.publish['index']['repos'] = {'dev': {}}
+    config_file.save()
+
+    project_name = 'My.App'
+
+    with temp_dir.as_cwd():
+        result = hatch('new', project_name)
+        assert result.exit_code == 0, result.output
+
+    path = temp_dir / 'my-app'
+
+    with path.as_cwd():
+        result = hatch('publish', '--user', 'foo', '--auth', 'bar')
+
+    assert result.exit_code == 1, result.output
+    assert result.output == 'Hatch config field `publish.index.repos.dev` must define a `url` key\n'
+
+
 def test_missing_user(hatch, temp_dir):
     project_name = 'My.App'
 
@@ -181,7 +190,7 @@ def test_missing_auth(hatch, temp_dir):
     assert result.output == 'Missing required option: auth\n'
 
 
-def test_flags(hatch, temp_dir_cache, helpers, published_project_name):
+def test_flags(hatch, devpi, temp_dir_cache, helpers, published_project_name):
     with temp_dir_cache.as_cwd():
         result = hatch('new', published_project_name)
         assert result.exit_code == 0, result.output
@@ -200,7 +209,7 @@ def test_flags(hatch, temp_dir_cache, helpers, published_project_name):
         artifacts = list(build_directory.iterdir())
 
         result = hatch(
-            'publish', '--user', '__token__', '--auth', PUBLISHER_TOKEN, '--repo', 'https://test.pypi.org/legacy/'
+            'publish', '--repo', devpi.repo, '--user', devpi.user, '--auth', devpi.auth, '--ca-cert', devpi.ca_cert
         )
 
     assert result.exit_code == 0, result.output
@@ -210,15 +219,17 @@ def test_flags(hatch, temp_dir_cache, helpers, published_project_name):
         {artifacts[1].relative_to(path)} ... success
 
         [{published_project_name}]
-        https://test.pypi.org/project/{published_project_name}/{current_version}/
+        {devpi.repo}{published_project_name}/{current_version}/
         """
     )
 
 
-def test_plugin_config(hatch, temp_dir_cache, helpers, published_project_name, config_file):
-    config_file.model.publish['index']['user'] = '__token__'
-    config_file.model.publish['index']['auth'] = PUBLISHER_TOKEN
-    config_file.model.publish['index']['repo'] = 'test'
+def test_plugin_config(hatch, devpi, temp_dir_cache, helpers, published_project_name, config_file):
+    config_file.model.publish['index']['user'] = devpi.user
+    config_file.model.publish['index']['auth'] = devpi.auth
+    config_file.model.publish['index']['ca-cert'] = devpi.ca_cert
+    config_file.model.publish['index']['repo'] = 'dev'
+    config_file.model.publish['index']['repos'] = {'dev': devpi.repo}
     config_file.save()
 
     with temp_dir_cache.as_cwd():
@@ -249,12 +260,60 @@ def test_plugin_config(hatch, temp_dir_cache, helpers, published_project_name, c
         {artifacts[1].relative_to(path)} ... success
 
         [{published_project_name}]
-        https://test.pypi.org/project/{published_project_name}/{current_version}/
+        {devpi.repo}{published_project_name}/{current_version}/
         """
     )
 
 
-def test_prompt(hatch, temp_dir_cache, helpers, published_project_name):
+def test_plugin_config_repo_override(hatch, devpi, temp_dir_cache, helpers, published_project_name, config_file):
+    config_file.model.publish['index']['user'] = 'foo'
+    config_file.model.publish['index']['auth'] = 'bar'
+    config_file.model.publish['index']['ca-cert'] = 'cert'
+    config_file.model.publish['index']['repo'] = 'dev'
+    config_file.model.publish['index']['repos'] = {
+        'dev': {'url': devpi.repo, 'user': devpi.user, 'auth': devpi.auth, 'ca-cert': devpi.ca_cert},
+    }
+    config_file.save()
+
+    with temp_dir_cache.as_cwd():
+        result = hatch('new', published_project_name)
+        assert result.exit_code == 0, result.output
+
+    path = temp_dir_cache / published_project_name
+
+    with path.as_cwd():
+        del os.environ[PublishEnvVars.REPO]
+
+        current_version = timestamp_to_version(helpers.get_current_timestamp())
+        result = hatch('version', current_version)
+        assert result.exit_code == 0, result.output
+
+        result = hatch('build')
+        assert result.exit_code == 0, result.output
+
+        build_directory = path / 'dist'
+        artifacts = list(build_directory.iterdir())
+
+        result = hatch('publish')
+
+    assert result.exit_code == 0, result.output
+    assert result.output == helpers.dedent(
+        f"""
+        {artifacts[0].relative_to(path)} ... success
+        {artifacts[1].relative_to(path)} ... success
+
+        [{published_project_name}]
+        {devpi.repo}{published_project_name}/{current_version}/
+        """
+    )
+
+
+def test_prompt(hatch, devpi, temp_dir_cache, helpers, published_project_name, config_file):
+    config_file.model.publish['index']['ca-cert'] = devpi.ca_cert
+    config_file.model.publish['index']['repo'] = 'dev'
+    config_file.model.publish['index']['repos'] = {'dev': devpi.repo}
+    config_file.save()
+
     with temp_dir_cache.as_cwd():
         result = hatch('new', published_project_name)
         assert result.exit_code == 0, result.output
@@ -272,11 +331,11 @@ def test_prompt(hatch, temp_dir_cache, helpers, published_project_name):
         build_directory = path / 'dist'
         artifacts = list(build_directory.iterdir())
 
-        result = hatch('publish', input='__token__\nfoo')
+        result = hatch('publish', input=f'{devpi.user}\nfoo')
 
     assert result.exit_code == 1, result.output
-    assert '403' in result.output
-    assert 'Invalid or non-existent authentication information' in result.output
+    assert '401' in result.output
+    assert 'Unauthorized' in result.output
 
     # Ensure nothing is saved for errors
     with path.as_cwd():
@@ -287,17 +346,17 @@ def test_prompt(hatch, temp_dir_cache, helpers, published_project_name):
 
     # Trigger save
     with path.as_cwd():
-        result = hatch('publish', str(artifacts[0]), input=f'__token__\n{PUBLISHER_TOKEN}')
+        result = hatch('publish', str(artifacts[0]), input=f'{devpi.user}\n{devpi.auth}')
 
     assert result.exit_code == 0, result.output
     assert result.output == helpers.dedent(
         f"""
-        Enter your username: __token__
-        Enter your credentials:{' '}
+        Username for '{devpi.repo}' [__token__]: {devpi.user}
+        Password / Token:{' '}
         {artifacts[0].relative_to(path)} ... success
 
         [{published_project_name}]
-        https://test.pypi.org/project/{published_project_name}/{current_version}/
+        {devpi.repo}{published_project_name}/{current_version}/
         """
     )
 
@@ -311,12 +370,67 @@ def test_prompt(hatch, temp_dir_cache, helpers, published_project_name):
         {artifacts[1].relative_to(path)} ... success
 
         [{published_project_name}]
-        https://test.pypi.org/project/{published_project_name}/{current_version}/
+        {devpi.repo}{published_project_name}/{current_version}/
         """
     )
 
 
-def test_external_artifact_path(hatch, temp_dir_cache, helpers, published_project_name):
+def test_initialize_auth(hatch, devpi, temp_dir_cache, helpers, published_project_name, config_file):
+    config_file.model.publish['index']['ca-cert'] = devpi.ca_cert
+    config_file.model.publish['index']['repo'] = 'dev'
+    config_file.model.publish['index']['repos'] = {'dev': devpi.repo}
+    config_file.save()
+
+    with temp_dir_cache.as_cwd():
+        result = hatch('new', published_project_name)
+        assert result.exit_code == 0, result.output
+
+    path = temp_dir_cache / published_project_name
+
+    # Trigger save
+    with path.as_cwd():
+        result = hatch('publish', '--initialize-auth', input=f'{devpi.user}\n{devpi.auth}')
+
+    assert result.exit_code == 0, result.output
+    assert result.output == helpers.dedent(
+        f"""
+        Username for '{devpi.repo}' [__token__]: {devpi.user}
+        Password / Token:{' '}
+        """
+    )
+
+    with path.as_cwd():
+        current_version = timestamp_to_version(helpers.get_current_timestamp())
+        result = hatch('version', current_version)
+        assert result.exit_code == 0, result.output
+
+        result = hatch('build', '-t', 'wheel')
+        assert result.exit_code == 0, result.output
+
+        build_directory = path / 'dist'
+        artifacts = list(build_directory.iterdir())
+
+    # Use saved results
+    with path.as_cwd():
+        result = hatch('publish', str(artifacts[0]))
+
+    assert result.exit_code == 0, result.output
+    assert result.output == helpers.dedent(
+        f"""
+        {artifacts[0].relative_to(path)} ... success
+
+        [{published_project_name}]
+        {devpi.repo}{published_project_name}/{current_version}/
+        """
+    )
+
+
+def test_external_artifact_path(hatch, devpi, temp_dir_cache, helpers, published_project_name, config_file):
+    config_file.model.publish['index']['ca-cert'] = devpi.ca_cert
+    config_file.model.publish['index']['repo'] = 'dev'
+    config_file.model.publish['index']['repos'] = {'dev': devpi.repo}
+    config_file.save()
+
     with temp_dir_cache.as_cwd():
         result = hatch('new', published_project_name)
         assert result.exit_code == 0, result.output
@@ -340,9 +454,7 @@ def test_external_artifact_path(hatch, temp_dir_cache, helpers, published_projec
         internal_build_directory = path / 'dist'
         internal_artifacts = list(internal_build_directory.iterdir())
 
-        result = hatch(
-            'publish', '--user', '__token__', '--auth', PUBLISHER_TOKEN, 'dist', str(external_build_directory)
-        )
+        result = hatch('publish', '--user', devpi.user, '--auth', devpi.auth, 'dist', str(external_build_directory))
 
     assert result.exit_code == 0, result.output
     assert result.output == helpers.dedent(
@@ -351,12 +463,17 @@ def test_external_artifact_path(hatch, temp_dir_cache, helpers, published_projec
         {external_artifacts[0]} ... success
 
         [{published_project_name}]
-        https://test.pypi.org/project/{published_project_name}/{current_version}/
+        {devpi.repo}{published_project_name}/{current_version}/
         """
     )
 
 
-def test_already_exists(hatch, temp_dir_cache, helpers, published_project_name):
+def test_already_exists(hatch, devpi, temp_dir_cache, helpers, published_project_name, config_file):
+    config_file.model.publish['index']['ca-cert'] = devpi.ca_cert
+    config_file.model.publish['index']['repo'] = 'dev'
+    config_file.model.publish['index']['repos'] = {'dev': devpi.repo}
+    config_file.save()
+
     with temp_dir_cache.as_cwd():
         result = hatch('new', published_project_name)
         assert result.exit_code == 0, result.output
@@ -374,7 +491,7 @@ def test_already_exists(hatch, temp_dir_cache, helpers, published_project_name):
         build_directory = path / 'dist'
         artifacts = list(build_directory.iterdir())
 
-        result = hatch('publish', '--user', '__token__', '--auth', PUBLISHER_TOKEN)
+        result = hatch('publish', '--user', devpi.user, '--auth', devpi.auth)
 
     assert result.exit_code == 0, result.output
     assert result.output == helpers.dedent(
@@ -383,16 +500,12 @@ def test_already_exists(hatch, temp_dir_cache, helpers, published_project_name):
         {artifacts[1].relative_to(path)} ... success
 
         [{published_project_name}]
-        https://test.pypi.org/project/{published_project_name}/{current_version}/
+        {devpi.repo}{published_project_name}/{current_version}/
         """
     )
 
-    for _ in range(30 if running_in_ci() else 5):
-        wait_for_artifacts(published_project_name, *artifacts)
-        time.sleep(1)
-
     with path.as_cwd():
-        result = hatch('publish', '--user', '__token__', '--auth', PUBLISHER_TOKEN)
+        result = hatch('publish', '--user', devpi.user, '--auth', devpi.auth)
 
     assert result.exit_code == 0, result.output
     assert result.output == helpers.dedent(
@@ -415,7 +528,7 @@ def test_no_artifacts(hatch, temp_dir_cache, helpers, published_project_name):
         directory.mkdir()
         (directory / 'test.txt').touch()
 
-        result = hatch('publish', 'dir1', 'dir2', '--user', '__token__', '--auth', 'foo')
+        result = hatch('publish', 'dir1', 'dir2', '--user', 'foo', '--auth', 'bar')
 
     assert result.exit_code == 1, result.output
     assert result.output == helpers.dedent(
@@ -425,10 +538,12 @@ def test_no_artifacts(hatch, temp_dir_cache, helpers, published_project_name):
     )
 
 
-def test_enable_with_flag(hatch, temp_dir_cache, helpers, published_project_name, config_file):
-    config_file.model.publish['index']['user'] = '__token__'
-    config_file.model.publish['index']['auth'] = PUBLISHER_TOKEN
-    config_file.model.publish['index']['repo'] = 'test'
+def test_enable_with_flag(hatch, devpi, temp_dir_cache, helpers, published_project_name, config_file):
+    config_file.model.publish['index']['user'] = devpi.user
+    config_file.model.publish['index']['auth'] = devpi.auth
+    config_file.model.publish['index']['ca-cert'] = devpi.ca_cert
+    config_file.model.publish['index']['repo'] = 'dev'
+    config_file.model.publish['index']['repos'] = {'dev': devpi.repo}
     config_file.model.publish['index']['disable'] = True
     config_file.save()
 
@@ -460,15 +575,17 @@ def test_enable_with_flag(hatch, temp_dir_cache, helpers, published_project_name
         {artifacts[1].relative_to(path)} ... success
 
         [{published_project_name}]
-        https://test.pypi.org/project/{published_project_name}/{current_version}/
+        {devpi.repo}{published_project_name}/{current_version}/
         """
     )
 
 
-def test_enable_with_prompt(hatch, temp_dir_cache, helpers, published_project_name, config_file):
-    config_file.model.publish['index']['user'] = '__token__'
-    config_file.model.publish['index']['auth'] = PUBLISHER_TOKEN
-    config_file.model.publish['index']['repo'] = 'test'
+def test_enable_with_prompt(hatch, devpi, temp_dir_cache, helpers, published_project_name, config_file):
+    config_file.model.publish['index']['user'] = devpi.user
+    config_file.model.publish['index']['auth'] = devpi.auth
+    config_file.model.publish['index']['ca-cert'] = devpi.ca_cert
+    config_file.model.publish['index']['repo'] = 'dev'
+    config_file.model.publish['index']['repos'] = {'dev': devpi.repo}
     config_file.model.publish['index']['disable'] = True
     config_file.save()
 
@@ -501,7 +618,7 @@ def test_enable_with_prompt(hatch, temp_dir_cache, helpers, published_project_na
         {artifacts[1].relative_to(path)} ... success
 
         [{published_project_name}]
-        https://test.pypi.org/project/{published_project_name}/{current_version}/
+        {devpi.repo}{published_project_name}/{current_version}/
         """
     )
 
@@ -529,16 +646,16 @@ class TestWheel:
         artifact_path = str(artifacts[0])
         metadata_file_path = f'{published_project_name}-{current_version}.dist-info/METADATA'
 
-        with zipfile.ZipFile(artifact_path, 'r') as zip_archive:
-            with zip_archive.open(metadata_file_path, 'r') as metadata_file:
-                metadata_file_contents = metadata_file.read().decode('utf-8')
+        with zipfile.ZipFile(artifact_path, 'r') as zip_archive, zip_archive.open(metadata_file_path) as metadata_file:
+            metadata_file_contents = metadata_file.read().decode('utf-8')
 
-        with zipfile.ZipFile(artifact_path, 'w') as zip_archive:
-            with zip_archive.open(metadata_file_path, 'w') as metadata_file:
-                metadata_file.write(remove_metadata_field(field, metadata_file_contents).encode('utf-8'))
+        with zipfile.ZipFile(artifact_path, 'w') as zip_archive, zip_archive.open(
+            metadata_file_path, 'w'
+        ) as metadata_file:
+            metadata_file.write(remove_metadata_field(field, metadata_file_contents).encode('utf-8'))
 
         with path.as_cwd():
-            result = hatch('publish', '--user', '__token__', '--auth', 'foo')
+            result = hatch('publish', '--user', 'foo', '--auth', 'bar')
 
         assert result.exit_code == 1, result.output
         assert result.output == helpers.dedent(
@@ -572,7 +689,7 @@ class TestSourceDistribution:
         extraction_directory = path / 'extraction'
 
         with tarfile.open(artifact_path, 'r:gz') as tar_archive:
-            tar_archive.extractall(extraction_directory)
+            tar_archive.extractall(extraction_directory, **helpers.tarfile_extraction_compat_options())
 
         metadata_file_path = extraction_directory / f'{published_project_name}-{current_version}' / 'PKG-INFO'
         metadata_file_path.write_text(remove_metadata_field(field, metadata_file_path.read_text()))
@@ -581,7 +698,7 @@ class TestSourceDistribution:
             tar_archive.add(extraction_directory, arcname='')
 
         with path.as_cwd():
-            result = hatch('publish', '--user', '__token__', '--auth', 'foo')
+            result = hatch('publish', '--user', 'foo', '--auth', 'bar')
 
         assert result.exit_code == 1, result.output
         assert result.output == helpers.dedent(
