@@ -11,14 +11,15 @@ def get_files(**kwargs):
     metadata_directory = kwargs.get('metadata_directory', '')
     package_paths = kwargs.get('package_paths', [])
 
-    files = []
-    for f in get_template_files(**kwargs):
-        if str(f.path) == 'LICENSE.txt':
-            files.append(File(Path(metadata_directory, 'licenses', f.path), f.contents))
+    files = [
+        File(Path(metadata_directory, 'licenses', f.path), f.contents)
+        for f in get_template_files(**kwargs)
+        if str(f.path) == 'LICENSE.txt'
+    ]
 
-    pth_file_name = f"{kwargs['package_name']}.pth"
-    files.append(File(Path(pth_file_name), '\n'.join(package_paths)))
-    files.append(
+    pth_file_name = f"_{kwargs['package_name']}.pth"
+    files.extend((
+        File(Path(pth_file_name), '\n'.join(package_paths)),
         File(
             Path(metadata_directory, 'WHEEL'),
             f"""\
@@ -28,18 +29,17 @@ Root-Is-Purelib: true
 Tag: py2-none-any
 Tag: py3-none-any
 """,
-        )
-    )
-    files.append(
+        ),
         File(
             Path(metadata_directory, 'METADATA'),
             f"""\
 Metadata-Version: {DEFAULT_METADATA_VERSION}
 Name: {kwargs['project_name']}
 Version: 0.0.1
+License-File: LICENSE.txt
 """,
-        )
-    )
+        ),
+    ))
 
     record_file = File(Path(metadata_directory, 'RECORD'), '')
     update_record_file_contents(record_file, files, generated_files={pth_file_name})
